# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import requests

from courseraresearchexports.constants.api_constants import \
    COURSE_API, PARTNER_API


def requests_response_to_model(response_transformer):
    """
    Creates decorator to handles errors in response from API call and
    transforms response with response_handler_func
    :param response_transformer: function(response) -> Any
    :return:
    """
    def response_transform_decorator(original_func):
        """
        Creates wrapper around a function that returns response
        """
        def response_transformer_wrapper(*args, **kwargs):
            """
            Log errors and apply transformation in response_handler_func
            """
            try:
                response = original_func(*args, **kwargs)
                response.raise_for_status()

            except requests.exceptions.HTTPError:
                logging.error(
                    'Request to {url} with body:\n\t{body}\nreceived response'
                    ':\n\t{text}\nPlease check request or contact '
                    'data-support@coursera.org if this error persists.'
                    .format(url=response.url,
                            text=response.text,
                            body=(response.request and response.request.body)))
                raise

            return response_transformer(response)
        return response_transformer_wrapper
    return response_transform_decorator


@requests_response_to_model(
    lambda response: response.json()['elements'][0]['slug'])
def lookup_course_slug_by_id(course_id):
    """
    Find the course slug given an course_id
    """
    return requests.get(requests.compat.urljoin(COURSE_API, course_id))


@requests_response_to_model(
    lambda response: response.json()['elements'][0]['id'])
def lookup_course_id_by_slug(course_slug):
    """
    Find the course_id given a course_slug
    """
    payload = {'q': 'slug', 'slug': course_slug}
    return requests.get(COURSE_API, params=payload)


@requests_response_to_model(
    lambda response: response.json()['elements'][0]['id'])
def lookup_partner_id_by_short_name(partner_short_name):
    """
    Find the partner_id by short name
    """
    payload = {'q': 'shortName', 'shortName': partner_short_name}
    return requests.get(COURSE_API, params=payload)


@requests_response_to_model(
    lambda response: response.json()['elements'][0]['shortName'])
def lookup_partner_short_name_by_id(partner_id):
    """
    Find the partner_id by short name
    """
    return requests.get(requests.compat.urljoin(PARTNER_API, str(partner_id)))



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Coursera's tools for managing docker containers configured with a
postgres database.
"""

import logging
import os
import shutil
import time

from courseraresearchexports import exports
from courseraresearchexports.models.ContainerInfo import ContainerInfo
from courseraresearchexports.constants.api_constants import \
    EXPORT_TYPE_TABLES
from courseraresearchexports.containers import utils as container_utils
from courseraresearchexports.exports import utils as export_utils


COURSERA_DOCKER_LABEL = 'courseraResearchExport'
COURSERA_LOCAL_FOLDER = os.path.expanduser('~/.coursera/exports/')
POSTGRES_DOCKER_IMAGE = 'postgres:9.5'
POSTGRES_INIT_MSG = 'PostgreSQL init process complete; ready for start up.'
POSTGRES_READY_MSG = 'database system is ready to accept connections'


def list_all(docker_client):
    """
    Return all containers that have Coursera label
    :param docker_client:
    :return containers_info: [ContainerInfo]
    """
    return [ContainerInfo.from_container(container['Id'], docker_client)
            for container in docker_client.containers(
            all=True, filters={'label': COURSERA_DOCKER_LABEL})]


def start(container_name, docker_client):
    """
    Start a docker container containing a research export database. Waits until
    """
    try:
        logging.debug('Starting container {}...'.format(container_name))
        docker_client.start(container_name)

        # poll logs to see if database is ready to accept connections
        while POSTGRES_READY_MSG not in docker_client.logs(
                container_name, tail=4):

            logging.debug('Polling container for database connection...')
            if not container_utils.is_container_running(
                    container_name, docker_client):
                raise RuntimeError('Container failed to start.')

            time.sleep(10)

        logging.info('Started container {}.'.format(container_name))

    except:
        logging.error(
            """Container failed to start, check log for errors:\n{}"""
            .format(docker_client.logs(container_name, tail=20)))
        raise


def stop(container_name, docker_client):
    """
    Stops a docker container
    """
    docker_client.stop(container_name)


def remove(container_name, docker_client):
    """
    Remove a stopped container
    """
    docker_client.remove_container(container_name)


def initialize(container_name, docker_client):
    """
    Initialize a docker container. Polls database for completion of
    entrypoint tasks.
    """
    try:
        logging.info('Initializing container {}...'.format(
            container_name))

        docker_client.start(container_name)
        while POSTGRES_INIT_MSG not in docker_client.logs(
                container_name, tail=20):

            logging.debug('Polling data for entrypoint initialization...')
            if not container_utils.is_container_running(container_name,
                                                        docker_client):
                raise RuntimeError('Container initialization failed.')

            time.sleep(10)

        logging.info('Initialized container {}.'.format(container_name))

    except:
        logging.error(
            """Container initialization failed, check log for errors:\n{}"""
            .format(docker_client.logs(container_name, tail=20)))
        logging.error(
            """If error persists, consider restarting your docker engine.""")
        raise


def create_from_folder(export_data_folder, docker_client,
                       container_name='coursera-exports',
                       database_name='coursera-exports'):
    """
    Using a folder containing a Coursera research export, create a docker
     container with the export data loaded into a data base and start the
     container
    :param export_data_folder: folder where export data/scripts is stored
    :param docker_client:
    :param container_name:
    :param database_name:
    :return container_id:
    """
    logging.debug('Creating containers from {folder}'.format(
        folder=export_data_folder))
    create_container_args = {
        'volumes': ['/mnt/exportData'],
        'host_config': docker_client.create_host_config(
            binds=['{}:/mnt/exportData:ro'.format(export_data_folder)],
            port_bindings={
                5432: container_utils.get_next_available_port(list_all(
                    docker_client))
            })
    }
    container = create_postgres_container(
        docker_client, container_name, database_name, create_container_args)

    container_id = container['Id']

    # copy containers initialization script to entrypoint
    database_setup_script = """
        createdb -U {user} {db}
        cd /mnt/exportData
        psql -e -U {user} -d {db} -f setup.sql
        psql -e -U {user} -d {db} -f load.sql
    """.format(user='postgres', db=database_name)

    docker_client.put_archive(
        container_id,  # using a named argument causes NullResource error
        path='/docker-entrypoint-initdb.d/',
        data=container_utils.create_tar_archive(
            database_setup_script, name='init-user-db.sh'))

    logging.info('Created container with id: {}'.format(container_id))

    initialize(container_id, docker_client)

    return container_id


def create_postgres_container(docker_client, container_name, database_name,
                              create_container_args):
    if not docker_client.images(name=POSTGRES_DOCKER_IMAGE):
        logging.info('Downloading image: {}'.format(POSTGRES_DOCKER_IMAGE))
        docker_client.import_image(image=POSTGRES_DOCKER_IMAGE)

    for existing_container in docker_client.containers(
            all=True, filters={'name': container_name}):
        logging.info('Removing existing container with name: {}'.format(
            container_name))
        docker_client.stop(existing_container)
        docker_client.remove_container(existing_container)
    create_container_args['image'] = POSTGRES_DOCKER_IMAGE
    create_container_args['name'] = container_name
    create_container_args['labels'] = {
        COURSERA_DOCKER_LABEL: None,
        'database_name': database_name
    }
    return docker_client.create_container(**create_container_args)


def create_from_export_request_id(export_request_id, docker_client,
                                  container_name=None, database_name=None):
    """
    Create a docker container containing the export data from a given
    export request. Container and database name will be inferred as the
    course slug or partner short name from export_request if not provided.
    :param export_request_id:
    :param docker_client:
    :param container_name:
    :param database_name:
    :return container_id:
    """
    export_request = exports.api.get(export_request_id)[0]

    if export_request.export_type != EXPORT_TYPE_TABLES:
        raise ValueError('Invalid Export Type. (Only tables exports supported.'
                         'Given [{}])'.format(export_request.export_type))

    logging.info('Downloading export {}'.format(export_request_id))
    downloaded_files = export_utils.download(
        export_request, dest=COURSERA_LOCAL_FOLDER)
    dest = os.path.join(COURSERA_LOCAL_FOLDER, export_request_id)
    for f in downloaded_files:
        container_utils.extract_zip_archive(
            archive=f,
            dest=dest,
            delete_archive=True)

    container_id = create_from_folder(
        export_data_folder=dest,
        docker_client=docker_client,
        database_name=(database_name if database_name
                       else export_request.scope_name),
        container_name=(container_name if container_name
                        else export_request.scope_name)
    )

    shutil.rmtree(dest)

    return container_id



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


RESEARCH_EXPORTS_APP = 'manage_research_exports'
RESEARCH_EXPORTS_API = 'https://www.coursera.org/api/onDemandExports.v2/'
COURSE_API = 'https://www.coursera.org/api/onDemandCourses.v1/'
PARTNER_API = 'https://www.coursera.org/api/partners.v1/'
CLICKSTREAM_API = 'https://www.coursera.org/api/clickstreamExportsDownload.v1/'
ANONYMITY_LEVEL_COORDINATOR = 'HASHED_IDS_NO_PII'
ANONYMITY_LEVEL_ISOLATED = 'HASHED_IDS_WITH_ISOLATED_UGC_NO_PII'
ANONYMITY_LEVELS = [ANONYMITY_LEVEL_COORDINATOR, ANONYMITY_LEVEL_ISOLATED]
EXPORT_TYPE_TABLES = 'RESEARCH_WITH_SCHEMAS'
EXPORT_TYPE_CLICKSTREAM = 'RESEARCH_EVENTING'
EXPORT_TYPE_GRADEBOOK = 'GRADEBOOK'
EXPORT_TYPES = [EXPORT_TYPE_TABLES, EXPORT_TYPE_CLICKSTREAM,
                EXPORT_TYPE_GRADEBOOK]
SCHEMA_NAMES = [
    'demographics',
    'users',
    'course_membership',
    'course_progress',
    'feedback',
    'assessments',
    'course_grades',
    'peer_assignments',
    'discussions',
    'programming_assignments',
    'course_content']



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from tabulate import tabulate

from courseraresearchexports.containers import client
from courseraresearchexports.containers import utils


def create_container(args):
    """
    Create a container containing a postgres database using an export job id.
    Export job will be downloaded and loaded into dockerized database.
    Automatically starts container.
    """
    d = utils.docker_client(args.docker_url, args.timeout)

    kwargs = {}
    if args.container_name:
        kwargs['container_name'] = args.container_name
    if args.database_name:
        kwargs['database_name'] = args.database_name

    if args.export_request_id:
        container_id = client.create_from_export_request_id(
            args.export_request_id, docker_client=d, **kwargs)
    elif args.export_data_folder:
        container_id = client.create_from_folder(
            args.export_data_folder, docker_client=d, **kwargs)

    logging.info('Container {:.12} ready.'.format(container_id))


def list_containers(args):
    """
    List docker containers created with Coursera data exports.
    """
    d = utils.docker_client(args.docker_url, args.timeout)
    containers_info = client.list_all(docker_client=d)

    if containers_info:
        containers_info_table = [['Name', 'Container Id', 'Database',
                                  'Created', 'Status', 'Host IP', 'Port']]

        for container_info in containers_info:
            containers_info_table.append([
                container_info.name,
                container_info.short_id,
                container_info.database_name,
                container_info.creation_time.strftime('%c'),
                container_info.status,
                container_info.host_ip,
                container_info.host_port
            ])

        logging.info('\n' + tabulate(containers_info_table,
                                     headers='firstrow'))


def start_container(args):
    """
    Start a docker container.
    """
    d = utils.docker_client(args.docker_url, args.timeout)
    client.start(args.container_name, docker_client=d)


def stop_container(args):
    """
    Stop a docker container.
    """
    d = utils.docker_client(args.docker_url, args.timeout)
    client.stop(args.container_name, docker_client=d)


def remove_container(args):
    """
    Remove a docker container, stop the container
    before removing.
    """
    d = utils.docker_client(args.docker_url, args.timeout)
    client.remove(args.container_name, docker_client=d)


def parser(subparsers):
    parser_containers = subparsers.add_parser(
        'containers',
        help='Create docker container from export jobs',
        description='Command line tools for creating a docker container'
        'containing the results of a research export. Please first '
        'authenticate with the OAuth2 client before making requests ('
        'courseraoauth2client config authorize --app manage-research-exports)',
        epilog='Please file bugs on github at: '
        'https://github.com/coursera/courseraresearchexports/issues. If you '
        'would like to contribute to this tool\'s development, check us out '
        'at: https://github.com/coursera/courseraresarchexports',
        parents=[utils.docker_client_arg_parser()])

    containers_subparsers = parser_containers.add_subparsers()

    parser_create = containers_subparsers.add_parser(
        'create',
        help=create_container.__doc__,
        description=create_container.__doc__)
    parser_create.set_defaults(func=create_container)

    source_subparser = parser_create.add_mutually_exclusive_group(
        required=True)

    source_subparser.add_argument(
        '--export_request_id',
        help='Export job to download and create containers')
    source_subparser.add_argument(
        '--export_data_folder',
        help='Location of already downloaded export data')

    parser_create.add_argument(
        '--container_name',
        help='Name for docker container.')
    parser_create.add_argument(
        '--database_name',
        help='Name for database inside container.')

    parser_list = containers_subparsers.add_parser(
        'list',
        help=list_containers.__doc__)
    parser_list.set_defaults(func=list_containers)

    parser_stop = containers_subparsers.add_parser(
        'stop',
        help=stop_container.__doc__)
    parser_stop.add_argument(
        'container_name',
        help='Name of the container to stop.')
    parser_stop.set_defaults(func=stop_container)

    parser_start = containers_subparsers.add_parser(
        'start',
        help=start_container.__doc__)
    parser_start.add_argument(
        'container_name',
        help='Name of the container to start.')
    parser_start.set_defaults(func=start_container)

    parser_remove = containers_subparsers.add_parser(
        'remove',
        help=remove_container.__doc__)
    parser_remove.add_argument(
        'container_name',
        help='Name of the container to remove.')
    parser_remove.set_defaults(func=remove_container)

    return parser_containers



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from tabulate import tabulate

import courseraresearchexports.db.db as db
from courseraresearchexports.containers import utils


def list_tables(args):
    """
    List all of the tables present in a dockerized database.
    """
    d = utils.docker_client(args.docker_url, args.timeout)
    tables = db.get_table_names(args.container_name, docker_client=d)
    logging.info('\n' + tabulate([[table] for table in tables]))


def list_views(args):
    """
    List all of the views present in a dockerized database.
    """
    d = utils.docker_client(args.docker_url, args.timeout)
    tables = db.get_view_names(args.container_name, docker_client=d)
    logging.info('\n' + tabulate([[table] for table in tables]))


def create_view(args):
    """
    Create a view from a sql query.
    """
    d = utils.docker_client(args.docker_url, args.timeout)

    if args.view_name:
        created_view = db.create_registered_view(
            args.container_name, args.view_name,
            args.partner_short_name, d)
    elif args.sql_file:
        created_view = db.create_view_from_file(
            args.container_name, args.sql_file,
            args.partner_short_name, d)

    logging.info('Created view {}'.format(created_view))


def unload_relation(args):
    """
    Unload a table or view to a CSV file.
    """
    d = utils.docker_client(args.docker_url, args.timeout)
    rowcount = db.unload_relation(args.container_name, args.dest,
                                  args.relation, d)

    logging.info('Unloaded {} rows'.format(rowcount))


def parser(subparsers):
    """Build an argparse argument parser to parse the command line."""

    # create the parser for the version subcommand.
    parser_db = subparsers.add_parser(
        'db',
        help='Tools for interacting with dockerized database',
        parents=[utils.docker_client_arg_parser()])

    db_subparsers = parser_db.add_subparsers()

    parser_tables = db_subparsers.add_parser(
        'list_tables',
        help=list_tables.__doc__)
    parser_tables.set_defaults(func=list_tables)
    parser_tables.add_argument(
        'container_name',
        help='Name of the container database.')

    parser_views = db_subparsers.add_parser(
        'list_views',
        help=list_views.__doc__)
    parser_views.set_defaults(func=list_views)
    parser_views.add_argument(
        'container_name',
        help='Name of the container database.')

    parser_create_view = db_subparsers.add_parser(
        'create_view',
        help=create_view.__doc__)
    parser_create_view.set_defaults(func=create_view)
    parser_create_view.add_argument(
        'container_name',
        help='Name of the container database.')
    create_source_subparser = parser_create_view.add_mutually_exclusive_group(
        required=True)
    create_source_subparser.add_argument(
        '--view_name',
        help='Name of view')
    create_source_subparser.add_argument(
        '--sql_file',
        help='SQL file with query.')
    parser_create_view.add_argument(
        '--partner_short_name',
        help='Your partner short name.')

    parser_unload = db_subparsers.add_parser(
        'unload_to_csv',
        help=unload_relation.__doc__)
    parser_unload.set_defaults(func=unload_relation)
    parser_unload.add_argument(
        'container_name',
        help='Name of the container database.')
    parser_unload.add_argument(
        '--dest',
        help='Destination folder.')
    parser_unload.add_argument(
        '--relation',
        help='Table or view to export.')

    return parser_db



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from courseraresearchexports.constants.api_constants import \
    ANONYMITY_LEVEL_COORDINATOR, ANONYMITY_LEVEL_ISOLATED, EXPORT_TYPE_TABLES,\
    EXPORT_TYPE_CLICKSTREAM, EXPORT_TYPE_GRADEBOOK, SCHEMA_NAMES
from courseraresearchexports.models import utils


class ExportRequest:
    """
    Represents a export request for Coursera's research data export
    service and provides methods for serialization.
    """

    def __init__(self, course_id=None, partner_id=None, group_id=None,
                 export_type=None, anonymity_level=None,
                 statement_of_purpose=None, schema_names=None,
                 interval=None, ignore_existing=None, **kwargs):
        self._course_id = course_id
        self._partner_id = partner_id
        self._group_id = group_id
        self._export_type = export_type
        self._anonymity_level = anonymity_level
        self._statement_of_purpose = statement_of_purpose
        self._schema_names = schema_names
        self._interval = interval
        self._ignore_existing = ignore_existing

    def to_json(self):
        """
        Serialize ExportRequest to a dictionary representing a json object.
        No validation is done with the exception that only specification of
        scope is used (course/partner/group).
        :return json_request:
        """
        json_request = {}

        if self._course_id:
            json_request['scope'] = {
                'typeName': 'courseContext',
                'definition': {
                    'courseId': self._course_id
                }}
        elif self._partner_id:
            json_request['scope'] = {
                'typeName': 'partnerContext',
                'definition': {
                    'partnerId': {
                        'maestroId': self._partner_id
                    }}}
        elif self._group_id:
            json_request['scope'] = {
                'typeName': 'groupContext',
                'definition': {
                    'groupId': self._group_id
                }}
        if self._export_type:
            json_request['exportType'] = self._export_type
        if self._anonymity_level:
            json_request['anonymityLevel'] = self._anonymity_level
        if self._statement_of_purpose:
            json_request['statementOfPurpose'] = self._statement_of_purpose
        if self._schema_names:
            json_request['schemaNames'] = self._schema_names
        if self._interval:
            json_request['interval'] = {
                'start': self._interval[0], 'end': self._interval[1]}
        if self._ignore_existing:
            json_request['ignoreExisting'] = self._ignore_existing

        return json_request

    @classmethod
    def from_args(cls, **kwargs):
        """
        Create a ExportResource object using the parameters required. Performs
        course_id/partner_id inference if possible.
        :param kwargs:
        :return export_request: ExportRequest
        """
        if kwargs.get('course_slug') and not kwargs.get('course_id'):
            kwargs['course_id'] = utils.lookup_course_id_by_slug(
                kwargs['course_slug'])
        elif kwargs.get('partner_short_name') and not kwargs.get('partner_id'):
            kwargs['partner_id'] = utils.lookup_partner_id_by_short_name(
                kwargs['partner_short_name'])

        if kwargs.get('user_id_hashing'):
            if kwargs['user_id_hashing'] == 'linked':
                kwargs['anonymity_level'] = ANONYMITY_LEVEL_COORDINATOR
            elif kwargs['user_id_hashing'] == 'isolated':
                kwargs['anonymity_level'] = ANONYMITY_LEVEL_ISOLATED

        return cls(**kwargs)

    @classmethod
    def from_json(cls, json_request):
        """
        Deserialize ExportRequest from json object.
        :param json_request:
        :return export_request: ExportRequest
        """
        kwargs = {}
        request_scope = json_request['scope']
        request_scope_context = request_scope['typeName']

        if request_scope_context == 'courseContext':
            kwargs['course_id'] = request_scope['definition']['courseId']
        elif request_scope_context == 'partnerContext':
            kwargs['partner_id'] = \
                request_scope['definition']['partnerId']['maestroId']
        elif request_scope_context == 'groupContext':
            kwargs['group_id'] = request_scope['definition']['groupId']

        if json_request.get('interval'):
            kwargs['interval'] = [
                json_request['interval']['start'],
                json_request['interval']['end']
            ]

        return cls(
            export_type=json_request.get('exportType'),
            anonymity_level=json_request.get('anonymityLevel'),
            statement_of_purpose=json_request.get('statementOfPurpose'),
            schema_names=json_request.get('schemaNames'),
            ignore_existing=json_request.get('ignoreExisting'),
            **kwargs)

    @property
    def course_id(self):
        return self._course_id

    @property
    def partner_id(self):
        return self._partner_id

    @property
    def export_type(self):
        return self._export_type

    @property
    def export_type_display(self):
        if self._export_type == EXPORT_TYPE_GRADEBOOK:
            return 'GRADEBOOK'
        elif self._export_type == EXPORT_TYPE_CLICKSTREAM:
            return 'CLICKSTREAM'
        elif self._export_type == EXPORT_TYPE_TABLES:
            return 'TABLES'
        else:
            return self._export_type

    @property
    def anonymity_level(self):
        return self._anonymity_level

    @property
    def formatted_anonymity_level(self):
        if self.anonymity_level == ANONYMITY_LEVEL_COORDINATOR:
            return 'Linked'
        elif self.anonymity_level == ANONYMITY_LEVEL_ISOLATED:
            return 'Isolated'
        else:
            return 'Unknown'

    @property
    def statement_of_purpose(self):
        return self._statement_of_purpose

    @property
    def interval(self):
        return self._interval

    @property
    def ignore_existing(self):
        return self._ignore_existing

    @property
    def scope_context(self):
        """
        Context for this ExportRequest, assume that only one identifier for
        partner/course/group is defined.
        """
        if self._course_id:
            return 'COURSE'
        elif self._partner_id:
            return 'PARTNER'
        elif self._group_id:
            return 'GROUP'
        else:
            return None

    @property
    def scope_id(self):
        """
        Identifier for the scope, assume that only one of course/partner/group
        is defined for a valid request.
        :return scope_id:
        """
        return self._course_id or self._partner_id or self._group_id

    @property
    def scope_name(self):
        """
        Human readable name for this scope context. course slugs for courses,
        partner short names for partners, but only group ids for groups (api is
        not open)
        :return:
        """
        if self._course_id:
            return utils.lookup_course_slug_by_id(self._course_id)
        elif self._partner_id:
            return utils.lookup_partner_short_name_by_id(self._partner_id)
        elif self._group_id:
            return self._group_id
        else:
            return 'UNKNOWN'

    @property
    def schema_names(self):
        return self._schema_names

    @property
    def schema_names_display(self):
        """
        Display only property for schemas names.
        :return schemas:
        """
        if self._schema_names:
            if set(self._schema_names) == set(SCHEMA_NAMES):
                return 'all'
            else:
                return ','.join(self._schema_names)
        else:
            return None

    def __eq__(self, other):
        """
        Override for internal equality checks as suggested at:
        http://stackoverflow.com/a/390640
        """
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from urlparse import urlparse

from tqdm import tqdm
import requests

from courseraresearchexports.constants.api_constants import \
    EXPORT_TYPE_CLICKSTREAM, EXPORT_TYPE_TABLES

from courseraresearchexports.exports import api
from courseraresearchexports.models.ClickstreamDownloadLinksRequest import \
    ClickstreamDownloadLinksRequest


def download(export_request, dest):
    """
    Download a data export job using a request id.
    """
    try:
        is_table_export = export_request.export_type == EXPORT_TYPE_TABLES
        is_clickstream_export = \
            export_request.export_type == EXPORT_TYPE_CLICKSTREAM

        if not export_request.download_link:
            if export_request.status in ['PENDING', 'IN_PROGRESS']:
                logging.error(
                    'Export request {} is currently {} and is not ready for'
                    'download. Please wait until the request is completed.'
                    .format(export_request.id, export_request.status))
                raise ValueError(
                    'Export request is not yet ready for download')
            elif export_request.status == 'TERMINATED':
                logging.error(
                    'Export request has been TERMINATED. Please contact '
                    'data-support@coursera.org if we have not resolved this '
                    'within 24 hours.')
                raise ValueError('Export request has been TERMINATED')
            elif is_clickstream_export:
                # We don't fill in download links for clickstream exports
                pass
            else:
                logging.error('Download link was not found.')
                raise ValueError('Download link was not found')

        if not os.path.exists(dest):
            logging.info('Creating destination folder: {}'.format(dest))
            os.makedirs(dest)

        if is_table_export:
            return [download_url(export_request.download_link, dest)]
        elif is_clickstream_export:
            links_request = ClickstreamDownloadLinksRequest.from_args(
                course_id=export_request.course_id,
                partner_id=export_request.partner_id,
                interval=export_request.interval)
            download_links = api.get_clickstream_download_links(links_request)
            return [download_url(link, dest) for link in download_links]
        else:
            raise ValueError('Require export_type is one of {} or {}'.format(
                EXPORT_TYPE_TABLES,
                EXPORT_TYPE_CLICKSTREAM))

    except Exception as err:
        logging.error('Download failed.\n{err}'.format(err=err))
        raise


def download_url(url, dest_folder):
    """
    Download url to dest_folder/FILENAME, where FILENAME is the last
    part of the url path.
    """
    filename = urlparse(url).path.split('/')[-1]
    full_filename = os.path.join(dest_folder, filename)
    response = requests.get(url, stream=True)
    chunk_size = 1024 * 1024
    logging.debug('Writing to file: {}'.format(full_filename))

    with open(full_filename, 'wb') as f:
        for data in tqdm(
                iterable=response.iter_content(chunk_size),
                total=int(response.headers['Content-length']) / chunk_size,
                unit='MB',
                desc=filename):
            f.write(data)
    return full_filename



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Coursera's wrapper for data exports API.
"""

import requests
from courseraoauth2client import oauth2
from courseraresearchexports.models.utils import requests_response_to_model
from courseraresearchexports.constants.api_constants import \
    RESEARCH_EXPORTS_APP, RESEARCH_EXPORTS_API, CLICKSTREAM_API
from courseraresearchexports.models.ExportRequestWithMetadata import \
    ExportRequestWithMetadata


@requests_response_to_model(ExportRequestWithMetadata.from_response)
def get(export_job_id):
    """
    Use Coursera's Research Export Resource to get a data export job given an
    export job id.
    :param export_job_id:
    :return export_request_with_metadata: [ExportRequestWithMetaData]
    """
    auth = oauth2.build_oauth2(app=RESEARCH_EXPORTS_APP).build_authorizer()
    response = requests.get(
        url=requests.compat.urljoin(RESEARCH_EXPORTS_API, export_job_id),
        auth=auth)

    return response


@requests_response_to_model(ExportRequestWithMetadata.from_response)
def get_all():
    """
    Uses Coursera's Research Exports Resource to get all data export job
    requests created by a user. Limited to the 100 most recent requests.
    :return export_requests: [ExportRequestWithMetaData]
    """
    auth = oauth2.build_oauth2(app=RESEARCH_EXPORTS_APP).build_authorizer()
    response = requests.get(
        url=RESEARCH_EXPORTS_API,
        auth=auth,
        params={'q': 'my'})

    return response


@requests_response_to_model(ExportRequestWithMetadata.from_response)
def post(export_request):
    """
    Creates a data export job using a formatted json request.
    :param export_request:
    :return export_request_with_metadata: [ExportRequestWithMetadata]
    """
    auth = oauth2.build_oauth2(app=RESEARCH_EXPORTS_APP).build_authorizer()
    response = requests.post(
        url=RESEARCH_EXPORTS_API,
        json=export_request.to_json(),
        auth=auth)

    return response


@requests_response_to_model(lambda response: response.json())
def get_clickstream_download_links(clickstream_download_links_request):
    """
    Return the download links for clickstream exports in a given scope.
    :param clickstream_download_links_request: ClickstreamDownloadLinksRequest
    """
    auth = oauth2.build_oauth2(app=RESEARCH_EXPORTS_APP).build_authorizer()
    response = requests.post(
        url=CLICKSTREAM_API,
        params=clickstream_download_links_request.to_url_params(),
        auth=auth)

    return response



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import pkg_resources

from courseraresearchexports.models.ExportDb import ExportDb


def get_table_names(container_name, docker_client):
    """
    Returns table names present in containerized database.
    :param container_name:
    :param docker_client:
    :return table_names:
    """
    export_db = ExportDb.from_container(container_name, docker_client)

    return export_db.tables


def get_view_names(container_name, docker_client):
    """
    Returns view names present in containerized database.
    :param container_name:
    :param docker_client:
    :return table_names:
    """
    export_db = ExportDb.from_container(container_name, docker_client)

    return export_db.views


def unload_relation(container_name, dest, relation, docker_client):
    """
    Unloads a table or view to a csv file.
    :param container_name:
    :param dest_file:
    :param relation:
    :param docker_client:
    :return:
    """
    if not os.path.exists(dest):
        logging.debug('Creating destination folder: {}'.format(dest))
        os.makedirs(dest)

    export_db = ExportDb.from_container(container_name, docker_client)
    output_filename = os.path.join(dest, '{}.csv'.format(relation))
    rowcount = export_db.unload_relation(relation, output_filename)
    return rowcount


def create_registered_view(container_name, view_name, partner_short_name,
                           docker_client):
    """
    Create a prepackaged view
    :param container_name:
    :param view_name:
    :param partner_short_name:
    :param docker_client:
    :return view_name:
    """
    export_db = ExportDb.from_container(container_name, docker_client)

    sql_text = pkg_resources.resource_string(
        __name__.split('.')[0], 'sql/{}.sql'.format(view_name))
    sql_text_with_partner_short_name = sql_text.replace(
        '[partner_short_name]', partner_short_name)

    export_db.create_view(view_name, sql_text_with_partner_short_name)

    return view_name


def create_view_from_file(container_name, sql_file, partner_short_name,
                          docker_client):
    """
    Create a view from a sql file.
    :param container_name:
    :param sql_file:
    :param partner_short_name:
    :param docker_client:
    :return view_name:
    """
    export_db = ExportDb.from_container(container_name, docker_client)

    with open(sql_file, 'r') as sf:
        sql_text = sf.read()

    view_name = os.path.split(os.path.basename(sql_text))[0]
    sql_text_with_partner_short_name = sql_text.replace(
        '[partner_short_name]', partner_short_name)

    export_db.create_view(view_name, sql_text_with_partner_short_name)

    return view_name



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dateutil.parser


class ContainerInfo:
    """
    Represents the relevant information about a docker container used to store
    a database of Coursera Export data.
    """

    def __init__(self, name=None, id=None, host_port=None, host_ip=None,
                 creation_time=None, database_name=None, status=None):
        self.name = name
        self.id = id
        self.short_id = id[:12] if id else None
        self.host_port = host_port
        self.host_ip = host_ip
        self.creation_time = creation_time
        self.status = status
        self.database_name = database_name

    @classmethod
    def from_container(cls, container_name, docker_client):
        """
        Create ContainerInfo using the response from docker-py Client's
        `inspect-container` method.
        :param container_dict:
        :return container_info: ContainerInfo
        """
        container_dict = docker_client.inspect_container(container_name)
        host_config = container_dict['HostConfig']['PortBindings']
        network_settings = container_dict['NetworkSettings']['Ports']

        assigned_port = int(host_config['5432/tcp'][0]['HostPort'])
        ip_if_running = network_settings and network_settings[
            '5432/tcp'][0]['HostIp']

        return cls(
            name=container_dict['Name'][1:],  # remove prepended '\'
            id=container_dict['Id'],
            creation_time=dateutil.parser.parse(container_dict['Created']),
            database_name=container_dict['Config']['Labels']['database_name'],
            status=container_dict['State']['Status'],
            host_port=assigned_port,
            host_ip=ip_if_running)



#!/usr/bin/env python

# PYTHON_ARGCOMPLETE_OK

# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Coursera's tools for interacting with research data exports.

You may install it from source, or via pip.
"""

import argcomplete
import argparse
import logging
import sys

from courseraresearchexports import commands
from courseraresearchexports.commands import utils


def build_parser():
    """
    Build an argparse argument parser to parse the command line.
    """

    parser = argparse.ArgumentParser(
        description="""Coursera tools for interacting with research exports.
        There are a number of subcommands, each with their own help
        documentation. Feel free to view them by executing `%(prog)s
        SUB_COMMAND -h`. For example: `%(prog)s jobs -h`.""",
        epilog="""Please file bugs on github at:
        https://github.com/coursera/courseraresearchexports/issues. If you
        would like to contribute to this tool's development, check us out at:
        https://github.com/coursera/courseraresarchexports""")

    utils.add_logging_parser(parser)

    # We have a number of subcommands. These subcommands have their own
    # subparsers. Each subcommand should set a default value for the 'func'
    # option. We then call the parsed 'func' function, and execution carries on
    # from there.
    subparsers = parser.add_subparsers()

    # create the parser for the version subcommand.
    commands.version.parser(subparsers)

    # create the parser for the jobs subcommand.
    commands.jobs.parser(subparsers)

    # create the parser for the containers subcommand.
    commands.containers.parser(subparsers)

    # create the parser for the db subcommand.
    commands.db.parser(subparsers)

    return parser


def main():
    """
    Boots up the command line tool
    """
    logging.captureWarnings(True)
    parser = build_parser()

    argcomplete.autocomplete(parser)

    args = parser.parse_args()
    # Configure logging
    args.setup_logging(args)
    # Dispatch into the appropriate subcommand function.
    try:
        return args.func(args)
    except SystemExit:
        raise
    except:
        logging.exception('Problem when running command. Sorry!')
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from io import BytesIO
import logging
import os
import tarfile
import time
import zipfile

from docker import Client


def extract_zip_archive(archive, dest, delete_archive=True):
    """
    Extracts a zip archive to `dest`
    :param export_archive:
    :param dest:
    :param delete_archive: delete the archive after extracting
    :return dest:
    """
    try:
        logging.debug('Extracting archive to {}'.format(dest))
        with zipfile.ZipFile(archive, 'r') as z:
            z.extractall(dest)
        if delete_archive:
            os.remove(archive)
    except:
        logging.error('Error in extracting zip archive {} to {}'.format(
            archive, dest))
        raise


def create_tar_archive(str, name='init-user-db.sh'):
    """
    Creates tar archive to load single file as suggested by
    https://gist.github.com/zbyte64/6800eae10ce082bb78f0b7a2cca5cbc2
    """
    archive_tarstream = BytesIO()
    archive_file = tarfile.TarFile(fileobj=archive_tarstream, mode='w')

    file_data = str.encode('utf8')
    file_info = tarfile.TarInfo(name)
    file_info.size = len(file_data)
    file_info.mtime = time.time()

    archive_file.addfile(file_info, BytesIO(file_data))
    archive_file.close()
    archive_tarstream.seek(0)

    return archive_tarstream


def get_next_available_port(containers_info):
    """
    Find next available port to map postgres port to host.
    :param containers_info:
    :return port:
    """
    ports = [container_info.host_port for container_info in containers_info]

    return (max(ports) + 1) if ports else 5433


def is_container_running(container_name, docker_client):
    """
    Check whether container is still running.
    :param container_name:
    :param docker_client:
    :return isRunning: Boolean
    """
    container_details = docker_client.inspect_container(container_name)

    return container_details['State']['Running']


def docker_client_arg_parser():
    """Builds an argparse parser for docker client connection flags."""
    # The following subcommands operate on a single containers. We centralize
    # all these options here.
    docker_parser = argparse.ArgumentParser(add_help=False)
    docker_parser.add_argument(
        '--docker-url',
        help='The url of the docker demon.')
    docker_parser.add_argument(
        '--timeout',
        type=int,
        default=60,
        help='Set the default timeout when interacting with the docker demon')
    return docker_parser


def docker_client(docker_url=None, timeout=60):
    """
    Attempts to create a docker client.

     - docker_url: base url for docker
     - timeout: timeout for docker client
     - returns: a docker-py client
    """
    if docker_url:
        return Client(
            base_url=docker_url,
            timeout=timeout,
            version='auto')
    else:
        return Client(
            timeout=timeout,
            version='auto')



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from courseraresearchexports.models import utils


class ClickstreamDownloadLinksRequest:
    """
    Represents a request for clickstream download links.
    """

    def __init__(self, course_id=None, partner_id=None, interval=None,
                 **kwargs):
        self.course_id = course_id
        self.partner_id = partner_id
        self.interval = interval

    @staticmethod
    def from_args(**kwargs):
        """
        Create a ClickstreamDownloadLinkRequest from arguments. Performs
        course_id/partner_id inference.
        :param kwargs:
        :return eventing_links_request: ClickstreamDownloadLinksRequest
        """
        if kwargs.get('course_slug') and not kwargs.get('course_id'):
            kwargs['course_id'] = utils.lookup_course_id_by_slug(
                kwargs['course_slug'])
        elif kwargs.get('partner_short_name') and not kwargs.get('partner_id'):
            kwargs['partner_id'] = \
                utils.lookup_partner_id_by_short_name(
                    kwargs['partner_short_name'])
        elif kwargs.get('group_id'):
            logging.error(
                'Eventing exports by group is not currently supported. '
                'Please see https://coursera.gitbooks.io/data-exports/content/'
                'introduction/programmatic_access.html')
            raise ValueError('Eventing exports by group is not supported.')

        return ClickstreamDownloadLinksRequest(**kwargs)

    @property
    def scope(self):
        """
        API specific format for request scope context.
        :return scope:
        """
        if self.course_id:
            return 'courseContext~{}'.format(self.course_id)
        elif self.partner_id:
            return 'partnerContext~{}'.format(self.partner_id)

    def to_url_params(self):
        """
        API specific parameters for POST request.
        :return:
        """
        url_params = {'action': 'generateLinks', 'scope': self.scope}
        if self.interval:
            url_params['startDate'] = self.interval[0]
            url_params['endDate'] = self.interval[1]

        return url_params



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Coursera's command line SDK for interacting with research data exports.

You may install it from source, or via pip.
"""

import sys
import logging


def command_version(args):
    """Implements the version subcommand"""

    # See http://stackoverflow.com/questions/17583443
    from pkg_resources import get_distribution, DistributionNotFound
    import os.path

    try:
        _dist = get_distribution('courseraresearchexports')
        # Normalize case for Windows systems
        dist_loc = os.path.normcase(_dist.location)
        here = os.path.normcase(__file__)
        if not here.startswith(
            os.path.join(
                dist_loc,
                'courseraresearchexports')):
            # not installed, but there is another version that *is*
            raise DistributionNotFound
    except DistributionNotFound:
        __version__ = 'Please install this project with setup.py'
    else:
        __version__ = _dist.version

    if args.quiet and args.quiet > 0:
        logging.info(__version__)
    else:
        logging.info("Your {prog}'s version is:\n\t{version}"
                     .format(prog=sys.argv[0], version=__version__))


def parser(subparsers):
    """Build an argparse argument parser to parse the command line."""

    # create the parser for the version subcommand.
    parser_version = subparsers.add_parser(
        'version',
        help="Output the version of %(prog)s to the console.")
    parser_version.set_defaults(func=command_version)

    return parser_version



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='courseraresearchexports',
    version='0.0.5',
    description='Command line tool for convenient access to '
    'Coursera Research Data Exports.',
    long_description=readme(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='coursera',
    url='https://github.com/coursera/courseraresearchexports',
    author='Timothy Lee',
    author_email='tlee@coursera.org',
    license='Apache',
    entry_points={
        'console_scripts': [
            'courseraresearchexports = courseraresearchexports.main:main',
        ],
    },
    packages=['courseraresearchexports',
              'courseraresearchexports.commands',
              'courseraresearchexports.constants',
              'courseraresearchexports.exports',
              'courseraresearchexports.containers',
              'courseraresearchexports.models',
              'courseraresearchexports.db'],
    package_data={'courseraresearchexports': ['sql/*.sql']},
    install_requires=[
        'argcomplete>=1.4.1',
        'courseraoauth2client>=0.0.1',
        'requests>=2.7.0,<2.11',
        'docker-py>=1.2.3',
        'tqdm>=4.8.4',
        'tabulate>=0.7.5',
        'python-dateutil>=2.5.3',
        'SQLAlchemy>=1.0.15'
    ],
    test_suite='nose.collector',
    tests_require=['nose', 'nose-cover3'],
    include_package_data=True,
    zip_safe=False
)



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tabulate import tabulate
import json
import argparse
import logging

from courseraresearchexports.exports import api
from courseraresearchexports.constants.api_constants import \
    ANONYMITY_LEVEL_COORDINATOR, EXPORT_TYPE_CLICKSTREAM, \
    EXPORT_TYPE_TABLES, SCHEMA_NAMES
from courseraresearchexports.models.ClickstreamDownloadLinksRequest import \
    ClickstreamDownloadLinksRequest
from courseraresearchexports.models.ExportRequest import ExportRequest
from courseraresearchexports.exports import utils


def request_clickstream(args):
    """
    Create and send an clickstream data export request with Coursera. Only
    available for data coordinators.
    """
    export_request = ExportRequest.from_args(
        course_id=args.course_id,
        course_slug=args.course_slug,
        partner_id=args.partner_id,
        parter_short_name=args.partner_short_name,
        group_id=args.group_id,
        anonymity_level=ANONYMITY_LEVEL_COORDINATOR,
        statement_of_purpose=args.purpose,
        export_type=EXPORT_TYPE_CLICKSTREAM,
        interval=args.interval,
        ignore_existing=args.ignore_existing)

    export_request_with_metadata = api.post(export_request)[0]

    logging.info('Successfully created clickstream export request {id}.'
                 .format(id=export_request_with_metadata.id))
    logging.debug('Request created with json body:\n{json}'
                  .format(json=json.dumps(
                        export_request_with_metadata.to_json(), indent=2)))


def request_tables(args):
    """
    Create and send a tables data export request with Coursera.
    """
    export_request = ExportRequest.from_args(
        course_id=args.course_id,
        course_slug=args.course_slug,
        partner_id=args.partner_id,
        parter_short_name=args.partner_short_name,
        group_id=args.group_id,
        user_id_hashing=args.user_id_hashing,
        statement_of_purpose=args.purpose,
        export_type=EXPORT_TYPE_TABLES,
        schema_names=args.schemas)

    export_request_with_metadata = api.post(export_request)[0]

    logging.info('Successfully created tables export request {id}.'
                 .format(id=export_request_with_metadata.id))
    logging.debug('Request created with json body:\n{json}'
                  .format(json=json.dumps(
                        export_request_with_metadata.to_json(), indent=2)))


def get(args):
    """
    Get the details and status of a data export request using a job id.
    """
    export_request = api.get(args.id)[0]

    export_request_info = [
        ['Export Job Id:', export_request.id],
        ['Export Type:', export_request.export_type_display],
        ['Status:', export_request.status],
        ['Scope Context:', export_request.scope_context],
        ['Scope Id:', export_request.scope_id],
        ['Scope Name:', export_request.scope_name],
        ['User id Hashing: ', export_request.formatted_anonymity_level],
        ['Created:', export_request.created_at.strftime('%c')]]

    if export_request.schema_names:
        export_request_info.append(
            ['Schemas:', export_request.schema_names_display])

    if export_request.download_link:
        export_request_info.append(
            ['Download Link:', export_request.download_link])

    if export_request.interval:
        export_request_info.append(
            ['Interval:', ' to '.join(export_request.interval)])

    logging.info('\n' + tabulate(export_request_info, tablefmt="plain"))


def get_all(args):
    """
    Get the details and status of your data export requests.
    """
    export_requests = api.get_all()

    export_requests_table = [['Created', 'Request Id', 'Status', 'Type',
                              'User Id Hashing', 'Scope', 'Schemas']]
    for export_request in sorted(export_requests, key=lambda x: x.created_at):
        export_requests_table.append([
            export_request.created_at.strftime('%Y-%m-%d %H:%M'),
            export_request.id,
            export_request.status,
            export_request.export_type_display,
            export_request.formatted_anonymity_level,
            export_request.scope_id,
            export_request.schema_names_display])

    logging.info('\n' + tabulate(export_requests_table, headers='firstrow'))


def download(args):
    """
    Download a data export job using a request id.
    """
    try:
        export_request = api.get(args.id)[0]
        dest = args.dest
        utils.download(export_request, dest)
    except Exception as err:
        logging.error('Download failed with exception:\n{}'.format(err))
        raise


def get_clickstream_links(args):
    """
    Generate links for clickstream data exports
    """
    clickstream_links_request = ClickstreamDownloadLinksRequest.from_args(
        course_id=args.course_id,
        course_slug=args.course_slug,
        partner_id=args.partner_id,
        parter_short_name=args.partner_short_name,
        group_id=args.group_id,
        interval=args.interval)

    clickstream_download_links = api.get_clickstream_download_links(
        clickstream_links_request)

    # TODO: add more descriptive information or option write to text file
    logging.info('\n' + tabulate(
        [[link] for link in clickstream_download_links],
        tablefmt="plain"))


def parser(subparsers):
    parser_jobs = subparsers.add_parser(
        'jobs',
        help='Get status of current/completed research export job(s)',
        description='Command line tools for requesting and reviewing the '
        'status of Coursera research data exports. Please first authenticate '
        'with the OAuth2 client before making requests (courseraoauth2client '
        'config authorize --app manage-research-exports).',
        epilog='Please file bugs on github at: '
        'https://github.com/coursera/courseraresearchexports/issues. If you '
        'would like to contribute to this tool\'s development, check us out '
        'at: https://github.com/coursera/courseraresarchexports')

    jobs_subparsers = parser_jobs.add_subparsers()

    create_request_parser(jobs_subparsers)

    parser_get_all = jobs_subparsers.add_parser(
        'get_all',
        help=get_all.__doc__,
        description=get_all.__doc__)
    parser_get_all.set_defaults(func=get_all)

    parser_get = jobs_subparsers.add_parser(
        'get',
        help=get.__doc__,
        description=get.__doc__)
    parser_get.set_defaults(func=get)

    parser_get.add_argument(
        'id',
        help='Export request ID')

    parser_download = jobs_subparsers.add_parser(
        'download',
        help=download.__doc__,
        description=download.__doc__)
    parser_download.set_defaults(func=download)

    parser_download.add_argument(
        'id',
        help='Export request ID')

    parser_download.add_argument(
        '--dest',
        default='.',
        help='Destination folder')

    parser_clickstream_links = jobs_subparsers.add_parser(
        'clickstream_download_links',
        help='Get download links for completed eventing exports.')
    parser_clickstream_links.set_defaults(func=get_clickstream_links)

    create_scope_subparser(parser_clickstream_links)

    parser_clickstream_links.add_argument(
        '--interval',
        nargs=2,
        metavar=('START', 'END'),
        help='Interval of exported clickstream data, inclusive. '
        '(i.e. 2016-08-01 2016-08-04).')

    return parser_jobs


def create_scope_subparser(parser):
    scope_subparser = parser.add_mutually_exclusive_group(
        required=True)
    scope_subparser.add_argument(
        '--course_id',
        help='Export rows corresponding to learners within a course according '
        'to the unique id assigned by Coursera.')
    scope_subparser.add_argument(
        '--course_slug',
        help='Export rows corresponding to learners within a course according '
        'to the unique name of your course defined as the part after '
        '/learn in the course url. (e.g. machine-learning for '
        'https://www.coursera.org/learn/machine-learning).')
    scope_subparser.add_argument(
        '--partner_id',
        help='Export rows corresponding to learners within a partner.')
    scope_subparser.add_argument(
        '--partner_short_name',
        help='Export rows corresponding to learners within a partner.')
    scope_subparser.add_argument(
        '--group_id',
        help='Export rows corresponding to learners without a group.')


def create_request_parser(subparsers):
    parser_request = subparsers.add_parser(
        'request',
        help='Create and send a data export request with Coursera.',
        description='Create and send a data export request with Coursera. '
        'Use subcommands to specify the export request type.')
    request_subparsers = parser_request.add_subparsers()

    # common arguments between schema and eventing exports
    request_args_parser = argparse.ArgumentParser(add_help=False)

    create_scope_subparser(request_args_parser)

    request_args_parser.add_argument(
        '--purpose',
        required=True,
        help='Please let us know how you plan to use the '
        'data, what types of research questions you\'re asking, who will '
        'be working with the data primarily, and with whom you plan to '
        'share it.')

    # tables subcommand
    parser_tables = request_subparsers.add_parser(
        'tables',
        help=request_tables.__doc__,
        description=request_tables.__doc__,
        parents=[request_args_parser])
    parser_tables.set_defaults(func=request_tables)

    parser_tables.add_argument(
        '--user_id_hashing',
        choices=['linked', 'isolated'],
        default='isolated',
        help='The level of user_id hashing in the data export. With \'linked\''
        ' user_id hashing, users can be identified between table schemas. '
        'With \'isolated\' user_id hashing, users have independent ids in'
        'different schemas and cannot be linked. Only data coordinators have '
        'access to \'linked\' users_ids to restrict PII.')

    parser_tables.add_argument(
        '--schemas',
        choices=SCHEMA_NAMES,
        nargs='+',
        default=SCHEMA_NAMES,
        help='Data schemas to export. Any combination of: {}. By default this '
        'will be all available schemas.'.format(
            ', '.join(SCHEMA_NAMES)))

    # clickstream subcommand
    parser_clickstream = request_subparsers.add_parser(
        'clickstream',
        help=request_clickstream.__doc__,
        description=request_clickstream.__doc__,
        parents=[request_args_parser])
    parser_clickstream.set_defaults(func=request_clickstream)

    parser_clickstream.add_argument(
        '--interval',
        nargs=2,
        metavar=('START', 'END'),
        help='Interval of clickstream data to be exported '
        '(i.e. 2016-08-01 2016-08-04). By default this will be the past day.')

    parser_clickstream.add_argument(
        '--ignore_existing',
        action='store_true',
        help='If flag is set, we will recompute clickstream data for all dates'
        'in the interval. Otherwise, previously computed days are skipped.')



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

import requests


def add_logging_parser(main_parser):
    """Build an argparse argument parser to parse the command line."""

    main_parser.set_defaults(setup_logging=set_logging_level)

    verbosity_group = main_parser.add_mutually_exclusive_group(required=False)
    verbosity_group.add_argument(
        '--verbose',
        '-v',
        action='count',
        help='Output more verbose logging. Can be specified multiple times.')
    verbosity_group.add_argument(
        '--quiet',
        '-q',
        action='count',
        help='Output less information to the console during operation. Can be '
        'specified multiple times.')

    main_parser.add_argument(
        '--silence-urllib3',
        action='store_true',
        help='Silence urllib3 warnings. See '
        'https://urllib3.readthedocs.org/en/latest/security.html for details.')

    return verbosity_group


def set_logging_level(args):
    """Computes and sets the logging level from the parsed arguments."""
    logging.basicConfig()
    root_logger = logging.getLogger()
    level = logging.INFO
    logging.getLogger('sqlalchemy.engine').setLevel(logging.WARNING)
    logging.getLogger('requests.packages.urllib3').setLevel(logging.WARNING)
    if "verbose" in args and args.verbose is not None:
        logging.getLogger('requests.packages.urllib3').setLevel(0)  # Unset
        if args.verbose > 1:
            level = 5  # "Trace" level
        elif args.verbose > 0:
            level = logging.DEBUG
        else:
            logging.critical("verbose is an unexpected value. {} exiting."
                             .format(args.verbose))
            sys.exit(2)
        logging.getLogger('sqlalchemy.engine').setLevel(level)
    elif "quiet" in args and args.quiet is not None:
        if args.quiet > 1:
            level = logging.ERROR
        elif args.quiet > 0:
            level = logging.WARNING
        else:
            logging.critical("quiet is an unexpected value. {} exiting."
                             .format(args.quiet))
    if level is not None:
        root_logger.setLevel(level)

    if args.silence_urllib3:
        # See: https://urllib3.readthedocs.org/en/latest/security.html
        requests.packages.urllib3.disable_warnings()



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv

from sqlalchemy import create_engine
from sqlalchemy.engine import reflection

from courseraresearchexports.models.ContainerInfo import ContainerInfo


class ExportDb:
    """
    Interface for accessing a database containing research export data.
    """
    def __init__(self, host_ip=None, host_port=None, db=None, **kwargs):

        if not (host_ip and host_port and db):
            raise ValueError(
                'Host IP, port and database name must be specified')

        self.host_ip = host_ip
        self.host_port = host_port
        self.db = db
        self.engine = create_engine(
            "postgresql://{user}@{host}:{port}/{db}"
            .format(user='postgres',
                    host=self.host_ip,
                    port=self.host_port,
                    db=self.db))

    @classmethod
    def from_container(cls, container_name, docker_client):
        """
        Create ExportDb object directly from container_name identifier.
        :param container_name:
        :param docker_client:
        :return:
        """
        container_info = ContainerInfo.from_container(container_name,
                                                      docker_client)
        return cls(host_ip=container_info.host_ip,
                   host_port=container_info.host_port,
                   db=container_info.database_name)

    def create_view(self, name, sql_text):
        """
        Creates or overrides an existing view given a select statement.
        :param name:
        :param sql_text:
        :return:
        """
        view_statement = """
        DROP VIEW IF EXISTS {name};
        CREATE VIEW {name} AS {sql_text};
        """.format(name=name, sql_text=sql_text)

        self.engine.execute(view_statement)

    def unload(self, query, output_filename):
        """
        Unloads to a csv file given a query.
        :param query:
        :param output_filename:
        :return rowcount:
        """
        result = self.engine.execute(query)

        rowcount = result.rowcount

        with open(output_filename, 'wb') as csv_file:
            csv_obj = csv.writer(csv_file)
            csv_obj.writerow(result.keys())
            for row in result:
                encoded_row = [col.encode('utf8')
                               if isinstance(col, unicode) else col
                               for col in row]
                csv_obj.writerow(encoded_row)

        return rowcount

    def unload_relation(self, relation, output_filename):
        """
        Unload a table or view.
        :param relation:
        :param output_filename:
        :return rowcount:
        """
        query = 'SELECT * FROM {relation};'.format(relation=relation)
        rowcount = self.unload(query, output_filename)
        return rowcount

    @property
    def tables(self):
        """
        Names of all tables present on database.
        """
        insp = reflection.Inspector.from_engine(self.engine)
        return insp.get_table_names()

    @property
    def views(self):
        """
        Names of all views present on database.
        """
        insp = reflection.Inspector.from_engine(self.engine)
        return insp.get_view_names()



# Copyright 2016 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import time

from courseraresearchexports.models.ExportRequest import ExportRequest


class ExportRequestMetadata:
    """Metadata about the internal timings of the export request"""

    def __init__(self, created_by=None, created_at=None, started_at=None,
                 completed_at=None, snapshot_at=None, **kwargs):
        self._created_by = created_by
        self._created_at = created_at
        self._started_at = started_at
        self._completed_at = completed_at
        self._snapshot_at = snapshot_at

    def to_json(self):
        """
        Serialize metadata from json object.
        :return json_metadata:
        """
        json_metadata = {}
        if self._created_by:
            json_metadata['createdBy'] = self._created_by
        if self._created_at:
            json_metadata['createdAt'] = datetime_to_unix_ms(self._created_at)
        if self._started_at:
            json_metadata['startedAt'] = datetime_to_unix_ms(self._started_at)
        if self._completed_at:
            json_metadata['completedAt'] = datetime_to_unix_ms(
                self._completed_at)
        if self._snapshot_at:
            json_metadata['snapshotAt'] = datetime_to_unix_ms(
                self._snapshot_at)

        return json_metadata

    @classmethod
    def from_json(cls, json_metadata):
        """
        Deserialize ExportRequestMetaData from json object.
        :param json_metadata:
        :return export_request_metadata: ExportRequestMetadata
        """
        if json_metadata:
            kwargs = {}
            if json_metadata.get('createdBy'):
                kwargs['created_by'] = json_metadata['createdBy']
            if json_metadata.get('createdAt'):
                kwargs['created_at'] = unix_ms_to_datetime(
                    json_metadata['createdAt'])
            if json_metadata.get('completedAt'):
                kwargs['completed_at'] = unix_ms_to_datetime(
                    json_metadata['completedAt'])
            if json_metadata.get('startedAt'):
                kwargs['started_at'] = unix_ms_to_datetime(
                    json_metadata['startedAt'])
            if json_metadata.get('snapshotAt'):
                kwargs['snapshot_at'] = unix_ms_to_datetime(
                    json_metadata['snapshotAt'])
            return cls(**kwargs)

        else:
            return None


class ExportRequestWithMetadata(ExportRequest):
    """
    Class representing a export request from Coursera's research data export
    service with metadata about its status.
    """

    def __init__(self, course_id=None, partner_id=None, group_id=None,
                 export_type=None, anonymity_level=None,
                 statement_of_purpose=None, schema_names=None,
                 interval=None, ignore_existing=None, id=None,
                 status=None, download_link=None, metadata=None, **kwargs):
        ExportRequest.__init__(
            self, course_id=course_id, partner_id=partner_id,
            group_id=group_id, export_type=export_type,
            anonymity_level=anonymity_level,
            statement_of_purpose=statement_of_purpose,
            schema_names=schema_names, interval=interval,
            ignore_existing=ignore_existing)
        self._id = id
        self._status = status
        self._download_link = download_link
        self._metadata = metadata

    def to_json(self):
        """
        Serialize ExportRequestWithMetadata to json object
        :return json_request:
        """
        json_request = ExportRequest.to_json(self)

        if self._id:
            json_request['id'] = self._id
        if self._status:
            json_request['status'] = self._status
        if self._download_link:
            json_request['downloadLink'] = self._download_link
        if self._metadata:
            json_request['metadata'] = self._metadata.to_json()

        return json_request

    @classmethod
    def from_export_request(cls, export_request, id=None, status=None,
                            download_link=None, metadata=None, **kwargs):
        """
        Create an object of class ExportRequestWithMetadata from an object of
        class ExportRequest.
        :param export_request: ExportRequest, parent object
        :param id:
        :param status:
        :param download_link:
        :param metadata:
        :param kwargs:
        :return export_request_with_metadata: ExportRequestWithMetadata
        """
        return cls(
            course_id=export_request._course_id,
            partner_id=export_request._partner_id,
            group_id=export_request._group_id,
            export_type=export_request._export_type,
            anonymity_level=export_request._anonymity_level,
            statement_of_purpose=export_request._statement_of_purpose,
            schema_names=export_request._schema_names,
            interval=export_request._interval,
            ignore_existing=export_request._ignore_existing,
            id=id,
            status=status,
            download_link=download_link,
            metadata=metadata)

    @classmethod
    def from_json(cls, json_request):
        """
        Deserialize ExportRequestWithMetadata from json object.
        :param json_request:
        :return export_request: ExportRequestWithMetadata
        """
        export_request = ExportRequest.from_json(json_request)

        return cls.from_export_request(
            export_request=export_request,
            id=json_request.get('id'),
            status=json_request.get('status'),
            download_link=json_request.get('downloadLink'),
            metadata=ExportRequestMetadata.from_json(
                json_request.get('metadata')))

    @classmethod
    def from_response(cls, response):
        """
        Instantiate a list of ExportRequestWithMeta objects from
        API call response.
        :param response:
        :return export_request_with_metadata_list: [ExportRequestWithMetadata]
        """
        return [cls.from_json(export_request)
                for export_request in response.json()['elements']]

    @property
    def id(self):
        return self._id

    @property
    def status(self):
        return self._status

    @property
    def download_link(self):
        return self._download_link

    @property
    def metadata(self):
        return self._metadata

    @property
    def created_at(self):
        if self._metadata and self._metadata._created_at:
            return self._metadata._created_at
        else:
            return datetime.fromtimestamp(0)


def datetime_to_unix_ms(dt):
    """Convert datetime object to timestamp in milliseconds"""
    return int(time.mktime(dt.timetuple()) * 1000)


def unix_ms_to_datetime(unix_ms):
    """Convert timestamp in milliseconds to datetime object"""
    return datetime.fromtimestamp(unix_ms / 1000.0)
